@echo off

REM ********************************************************************
REM * This is a sample batch file to demonstrate manually installing
REM * or removing the PLX PCI/PCIe Service driver.  It utilizes the
REM * Microsoft SC.EXE utility to access the Service Control Manager.
REM * This should work in Windows 2000 & higher.
REM *
REM * Refer to the Usage for additional details.
REM ********************************************************************



REM Determine if 32-bit or 64-bit Windows
SET b64Bit=0
IF EXIST %SystemRoot%\SysWOW64 SET b64Bit=1


REM Verify Admin priviledges
REM  Use FsUtil.exe (returns 1=error 0=ok) to test privileges
FSUTIL >NUL
IF ERRORLEVEL 1 goto _Error_NoAdmin


REM Verify command-line
If /i "%1" == "Install" goto _Install
If /i "%1" == "Remove"  goto _Remove
goto _ShowUsage


:_ShowUsage
echo.
echo PlxSvcCtrl - Script to install or remove the PLX PCI Service Driver
echo.
echo   Usage:  PlxSvcCtrl install ^| remove
echo.
echo   Where:  install  - Copies PlxSvc.sys to Windows drivers folder
echo                      Creates the PlxSvc service
echo                      Starts the driver
echo.
echo           remove   - Stops the driver
echo                      Delete the PlxSvc service
echo                      Delete the PlxSvc.sys file from Windows drivers folder
echo.
echo   Notes:
echo     This script requires administrator priviledges to execute or will fail.
echo.
echo     This script uses the Windows SC.EXE utility to make calls to the
echo     Service Control Manager (SCM), which is responsible for managing
echo     system services. SC allows creation, delete, control, etc. of services.
echo.
echo     The script expects 2 versions of PlxSvc.sys located in the following
echo     subdirectories from the script file:
echo.
echo      Driver_PlxSvc\i386\PlxSvc.sys  - PLX Service driver for 32-bit Windows
echo      Driver_PlxSvc\Amd64\PlxSvc.sys - PLX Service driver for 64-bit Windows
goto _Exit



:_Install
echo.
echo Install the PLX service driver
echo ======================================

REM Verify Service is not installed already (1060=Service not installed)
SC QUERYEX PlxSvc > NUL
IF NOT ERRORLEVEL 1060 goto _Error_SvcInstalledAlready

REM Verify driver doesn't already exist
IF EXIST %systemRoot%\System32\Drivers\PlxSvc.sys goto _Error_DriverExists

REM Verify correct version of PlxSvs,sys exists
IF %b64Bit% == 0 SET PLX_SVC_DIR=Driver_PlxSvc\i386
IF %b64Bit% == 1 SET PLX_SVC_DIR=Driver_PlxSvc\amd64
IF NOT EXIST %PLX_SVC_DIR%\PlxSvc.sys goto _Error_DriverNotFound

echo - Copy %PLX_SVC_DIR%\PlxSvc.sys --^> %systemRoot%\System32\Drivers...
Copy %PLX_SVC_DIR%\PlxSvc.sys %systemRoot%\System32\Drivers > NUL

echo - Request SCM to create the service...
SC CREATE PlxSvc binPath= System32\Drivers\PlxSvc.sys type= kernel ^
          start= auto error= ignore DisplayName= "PLX PCI/PCIe Service Driver" > NUL
IF ERRORLEVEL 1073 goto _Error_SvcInstalledAlready
IF ERRORLEVEL 1 goto _Error_SCM

echo - Add the Common Buffer Size option to registry...
REG ADD "HKLM\System\CurrentControlSet\services\PlxSvc" ^
    /v CommonBufferSize /t REG_DWORD /d 8192 > NUL
IF ERRORLEVEL 1 goto _Error_SCM

echo - Start the service...
SC START PlxSvc >NUL
IF ERRORLEVEL 1056 goto _Error_SvcAlreadyRunning
IF ERRORLEVEL 1 goto _Error_SCM

echo.
echo   -- PLX PCI/PCIe Service driver installed successfully --
goto _Exit



:_Remove
echo.
echo Remove the PLX service driver
echo ======================================

REM Verify Service is installed already (1060=Service not installed)
SC QUERYEX PlxSvc > NUL
IF ERRORLEVEL 1060 goto _Error_SvcNotInstalled

echo - Stop the service...
SC STOP PlxSvc >NUL
IF ERRORLEVEL 1062 ECHO    - SCM reports service already stopped

echo - Request SCM to delete the service...
SC DELETE PlxSvc > NUL
IF ERRORLEVEL 1060 goto _Error_SvcNotInstalled
IF ERRORLEVEL 1 goto _Error_SCM

echo - Delete %systemRoot%\System32\Drivers\PlxSvc.sys...
Del %systemRoot%\System32\Drivers\PlxSvc.sys > NUL

echo.
echo   -- PLX PCI/PCIe Service driver removed successfully --
goto _Exit



:_Error_NoAdmin
echo ERROR: This script must be executed with Administrator priviledges.
goto _Exit


:_Error_DriverExists
echo ERROR: PlxSvc.sys already exists in %systemRoot%\System32\Drivers
echo        Please delete this file and re-run the scipt.
goto _Exit


:_Error_DriverNotFound
echo ERROR: PlxSvc.sys not found in %PLX_SVC_DIR%
goto _Exit


:_Error_SvcInstalledAlready
echo ERROR: The SCM reports the PLX Service is already installed
goto _Exit


:_Error_SvcNotInstalled
echo ERROR: The SCM reports the PLX Service is not installed
goto _Exit


:_Error_SvcAlreadyRunning
echo ERROR: The SCM reports the PLX Service is already running
goto _Exit


:_Error_SCM
echo ERROR: The SCM reported error code %ERRORLEVEL%
goto _Exit


:_Exit
